/*
  Copyright (C) 2009 Jon Macey

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __SHADERMANAGER__
#define __SHADERMANAGER__
/// @file ShaderManager.h
/// @brief main shader loader / manager class for GLSL shaders
// must include types.h first for ngl::Real and GLEW if required
#include "Types.h"
#include "Singleton.h"
#include <cstdlib>
#include <iostream>
#include <string>
#include <map>
#include "Matrix.h"
#include "Vector.h"
#include "Shader.h"
#include "ShaderProgram.h"

//----------------------------------------------------------------------------------------------------------------------
// part of the python bindings
#ifdef NO_PYTHON_LIB
  #include <boost/python.hpp>
  #include <boost/format.hpp>
#endif

namespace ngl
{


//----------------------------------------------------------------------------------------------------------------------
///  @class ShaderManager  "include/ShaderManager.h"
///  @brief Singleton Class to init and Use GLSL Shaders
///  the class stores the shaders as a map of shader objects and they can be found by name
/// Updated to use templated singleton
/// Updated to use a map to store the shaders by name and hence removed the name from the ShaderObject Class
/// Also Added methods to adjust shader parameters by shader / param name, at present use only 1-4 Floats and
/// a few others need to think of a better way of doing this
///  @author Jonathan Macey
///  @version 3.0
///  @date Last Revision 20/04/10 used templated singleton
//----------------------------------------------------------------------------------------------------------------------


class NGL_DLLEXPORT ShaderManager : public  Singleton<ShaderManager>
{
  friend class Singleton<ShaderManager>;

public :

  //----------------------------------------------------------------------------------------------------------------------
  /// @brief create an empty ShaderProgram for us to attach shaders etc to
  /// @param _name the name of the ShaderProgram to link
  //----------------------------------------------------------------------------------------------------------------------
  void createShaderProgram(
                           std::string _name
                          );
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief attatch a Shader to the ShaderProgram referenced by _name
  /// @param _name the name of the ShaderProgram to attach
  //----------------------------------------------------------------------------------------------------------------------
  void attachShader(
                    std::string _name,
                    SHADERTYPE _type
                   );
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief attatch a Shader to the ShaderProgram referenced by _name
  /// @param _program the name of the ShaderProgram to attach to
  /// @param _shader the name of the Shader to attach to Program
  //----------------------------------------------------------------------------------------------------------------------

  void attachShaderToProgram(
                             std::string _program,
                             std::string _shader
                            );
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief get the Program ID of the GL Program by name
  /// @param _name the name of the ShaderProgram to find
  /// @returns the id of the program found or -1 on error
  //----------------------------------------------------------------------------------------------------------------------
  GLuint getProgramID(
                      std::string _name
                     );
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief compile the shader from _name
  /// @param _name the name of the ShaderProgram to compile
  //----------------------------------------------------------------------------------------------------------------------
  void compileShader(
                      std::string _name
                    );
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief link the program Object  from _name
  /// @param _name the name of the ShaderProgram to link
  //----------------------------------------------------------------------------------------------------------------------
  void linkProgramObject(
                          std::string _name
                        );
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief toggle debug mode
  //----------------------------------------------------------------------------------------------------------------------
  inline void toggleDebug() { m_debugState ^=true;}
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief set active shader to name (if not found sets glProgramObject(0)
  /// @param _name the name of the ShaderProgram to use
  //----------------------------------------------------------------------------------------------------------------------
  void use(
            std::string _name
          );

  void bindAttribute(
                      std::string _programName,
                      GLuint _index,
                      std::string _attribName
                    );

  ShaderProgram * operator[](const std::string &_programName);
  ShaderProgram * operator[](const char *_programName);

  //----------------------------------------------------------------------------------------------------------------------
  /// @brief method to load shaders
  /// @param[in] _shaderName the name of the shader to be stored in the Manager
  /// @param[in] _vert the vertex shader to load
  /// @param[in] _frag the fragment shader to load
  /// @param[in] _geo an optional geo shader to load
  /// @param[in] _exitOnError exit if there is an error in the shader
  //----------------------------------------------------------------------------------------------------------------------
  void loadShader(
                  const std::string &_shaderName,
                  const std::string &_vert,
                  const std::string &_frag,
                  const std::string &_geo="",
                  const bool _exitOnError=false
                 );


  //----------------------------------------------------------------------------------------------------------------------
  /// @brief se the shader passed in by name
  /// @param[in] _name the name of the shader to load
  //----------------------------------------------------------------------------------------------------------------------
  void useShader(
                 const std::string &_name
                );

  //----------------------------------------------------------------------------------------------------------------------
  /// @brief Toggle debug mode on
  //----------------------------------------------------------------------------------------------------------------------
  inline void debugOn()
                      {
                      m_debugState=true;
                      }

  //----------------------------------------------------------------------------------------------------------------------
  /// @brief Toggle debug mode off
  //----------------------------------------------------------------------------------------------------------------------
  inline void debugOff()
                       {
                       m_debugState=false;
                       }

  //----------------------------------------------------------------------------------------------------------------------
  /// @brief  get shader handle by index
  //----------------------------------------------------------------------------------------------------------------------
  GLhandleARB operator[](
                         const ShaderManager&
                        );

  //----------------------------------------------------------------------------------------------------------------------
  /// @brief get the number of shaders stored
  /// @returns the number of shaders
  //----------------------------------------------------------------------------------------------------------------------
  unsigned int getNumShaders()
                             {
                             return m_shaders.size();
                             }
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief set a shader param by name for 1 int param note that the shader
  /// must be the currently active shader of else this will fail
  /// @param[in] _shaderName the name of the shader to set the param for
  /// @param[in] _paramName the name of the parameter in the shader to set
  /// @param[in] _p1 the matrix to set from (float 16) value of the parameter to set
  //----------------------------------------------------------------------------------------------------------------------
  void setShaderParamFromMatrix(
                                const std::string &_shaderName,
                                const std::string &_paramName,
                                ngl::Matrix _p1
                               );

  //----------------------------------------------------------------------------------------------------------------------
  /// @brief set a shader param by name for 1 int param note that the shader
  /// must be the currently active shader of else this will fail
  /// @param[in] _shaderName the name of the shader to set the param for
  /// @param[in] _paramName the name of the parameter in the shader to set
  /// @param[in] _p1 the vector to set from (float 4) value of the parameter to set
  //----------------------------------------------------------------------------------------------------------------------
  void setShaderParamFromVector(
                                const std::string &_shaderName,
                                const std::string &_paramName,
                                ngl::Vector _p1
                               );

  //----------------------------------------------------------------------------------------------------------------------
  /// @brief set a shader param by name for 1 int param note that the shader
  /// must be the currently active shader of else this will fail
  /// @param[in] _shaderName the name of the shader to set the param for
  /// @param[in] _paramName the name of the parameter in the shader to set
  /// @param[in] _p1 the float value of the parameter to set
  //----------------------------------------------------------------------------------------------------------------------
  void setShaderParam1i(
                        const std::string &_shaderName,
                        const std::string &_paramName,
                        int _p1
                        );
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief set a shader param by name for 1 float param note that the shader
  /// must be the currently active shader of else this will fail
  /// @param[in] _shaderName the name of the shader to set the param for
  /// @param[in] _paramName the name of the parameter in the shader to set
  /// @param[in] _p1 the float value of the parameter to set
  //----------------------------------------------------------------------------------------------------------------------
  void setShaderParam1f(
                        const std::string &_shaderName,
                        const std::string &_paramName,
                        float _p1
                        );
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief set a shader param by name for 2 float params note that the shader
  /// must be the currently active shader of else this will fail
  /// @param[in] _shaderName the name of the shader to set the param for
  /// @param[in] _paramName the name of the parameter in the shader to set
  /// @param[in] _p1 the float value of the parameter to set
  /// @param[in] _p2 the float value of the parameter to set
  //----------------------------------------------------------------------------------------------------------------------
  void setShaderParam2f(
                        const std::string &_shaderName,
                        const std::string &_paramName,
                        float _p1,
                        float _p2
                        );
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief set a shader param by name for 3 float params note that the shader
  /// must be the currently active shader of else this will fail
  /// @param[in] _shaderName the name of the shader to set the param for
  /// @param[in] _paramName the name of the parameter in the shader to set
  /// @param[in] _p1 the float value of the parameter to set
  /// @param[in] _p2 the float value of the parameter to set
  /// @param[in] _p3 the float value of the parameter to set
  //----------------------------------------------------------------------------------------------------------------------
  void setShaderParam3f(
                        const std::string &_shaderName,
                        const std::string &_paramName,
                        float _p1,
                        float _p2,
                        float _p3
                        );
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief set a shader param by name for 4 float params note that the shader
  /// must be the currently active shader of else this will fail
  /// @param[in] _shaderName the name of the shader to set the param for
  /// @param[in] _paramName the name of the parameter in the shader to set
  /// @param[in] _p1 the float value of the parameter to set
  /// @param[in] _p2 the float value of the parameter to set
  /// @param[in] _p3 the float value of the parameter to set
  /// @param[in] _p4 the float value of the parameter to set
  //----------------------------------------------------------------------------------------------------------------------
  void setShaderParam4f(
                        const std::string &_shaderName,
                        const std::string &_paramName,
                        float _p1,
                        float _p2,
                        float _p3,
                        float _p4
                        );
  /// @brief reset the Shader manager which will delete all shader objects
  void reset();
  /// return the pointer to the shader attribute location
  GLint getAttribLocation(
                            const std::string &_shaderName,
                            const std::string &_paramName
                           );
	void loadShaderSource(
	                           std::string _shaderName,
	                           std::string _sourceFile
	                          );
  void useNullProgram();

protected:


	/// @brief a map of shader Programs using name as key to shader pointer
  std::map <std::string,ShaderProgram *> m_shaderPrograms;
  /// @brief map of shaders using name as key
  std::map <std::string,Shader *> m_shaders;
  /// @brief null ShaderProgram so we can return when shader not know;
  ShaderProgram *m_nullProgram;

  //----------------------------------------------------------------------------------------------------------------------
  /// @brief  flag to indicate the debug state
  //----------------------------------------------------------------------------------------------------------------------
  bool m_debugState;

  //----------------------------------------------------------------------------------------------------------------------
  /// @brief  the nunmber of shaders loaded
  //----------------------------------------------------------------------------------------------------------------------
  unsigned int m_numShaders;
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief default ctor private as a singleton
  //----------------------------------------------------------------------------------------------------------------------
  ShaderManager();
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief  copy ctor private as singleton
  //----------------------------------------------------------------------------------------------------------------------
  ShaderManager(
               const ShaderManager&
               );

	/// extra glue for python lib bindings nothing to see here (unless ....)
	#ifdef NO_PYTHON_LIB
	  public :
	    ShaderProgram *getShaderProgram(
																			const std::string &_programName
																		 );
	#endif


}; // end of class

} // end ngl Namespace

#endif
//----------------------------------------------------------------------------------------------------------------------

