/*
  Copyright (C) 2009 Jon Macey

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __VERTEXARRAYOBJECT_H__
#define __VERTEXARRAYOBJECT_H__
//----------------------------------------------------------------------------------------------------------------------
/// @file VertexArrayObject.h
/// @brief a class to store an OpenGL VAO
//----------------------------------------------------------------------------------------------------------------------
// part of the python bindings
#ifdef NO_PYTHON_LIB
  #include <boost/python.hpp>
  #include <boost/format.hpp>
#endif

// must include types.h first for ngl::Real and GLEW if required
#include "Types.h"


//----------------------------------------------------------------------------------------------------------------------
/// @file VertexArrayObject.h
/// @brief encapsulates an OpenGL VAO
/// @author Jonathan Macey
/// @version 1.0
/// @date 14/6/11 Initial version
/// Revision History :

//----------------------------------------------------------------------------------------------------------------------
namespace ngl
{

//----------------------------------------------------------------------------------------------------------------------
/// @class VertexArrayObject "include/VertexArrayObject.h"
/// @brief a class to encapsulate an OpenGL VAO
//----------------------------------------------------------------------------------------------------------------------


class NGL_DLLEXPORT VertexArrayObject
{

public:
	//----------------------------------------------------------------------------------------------------------------------
	/// @brief method to return a newVOA object
	/// @returns a new VOA object
	//----------------------------------------------------------------------------------------------------------------------
	inline static VertexArrayObject * createVOA(
																								GLenum _mode
																							)
																							 {
																								 return new VertexArrayObject(_mode);
																							 }
	//----------------------------------------------------------------------------------------------------------------------
	/// @brief bind this VOA to make it active
	//----------------------------------------------------------------------------------------------------------------------
	void bind();
	//----------------------------------------------------------------------------------------------------------------------
	/// @brief unbind the VOA and set 0 to be the active VOA
	//----------------------------------------------------------------------------------------------------------------------
	void unbind();
	//----------------------------------------------------------------------------------------------------------------------
	/// @brief delete the VOA (using OpenGL calls to clear data etc)
	//----------------------------------------------------------------------------------------------------------------------
	void removeVOA();
	//----------------------------------------------------------------------------------------------------------------------
	/// @brief accesor for the VOA id
	/// @brief returns the VOA id
	//----------------------------------------------------------------------------------------------------------------------
	inline GLuint getID()const {return m_id;}
	//----------------------------------------------------------------------------------------------------------------------
	/// @brief accesor to see if VOA is bound
	/// @brief returns the bind state (true for bound)
	//----------------------------------------------------------------------------------------------------------------------
	inline bool isBound()const {return m_bound;}
	//----------------------------------------------------------------------------------------------------------------------
	/// @brief accesor to see if VOA has allocated data
	/// @brief returns the allocated state (true for allocated)
	//----------------------------------------------------------------------------------------------------------------------
	inline bool isAllocated()const {return m_allocated;}
	//----------------------------------------------------------------------------------------------------------------------
	/// @brief allocate our data
	/// @param _size the size of the raw data passed
	/// @param _data the actual data to set for the VOA
	/// @param _mode the draw mode hint used by GL
	//----------------------------------------------------------------------------------------------------------------------
	void setData(
								unsigned int _size,
	              const GLfloat &_data,
							  GLenum _mode=GL_STATIC_DRAW
							);
	//----------------------------------------------------------------------------------------------------------------------
		/// @brief allocate our data
		/// @param _size the size of the raw data passed (not counting sizeof(GL_FLOAT))
		/// @param _type the data type of the Pointer (eg GL_FLOAT)
		/// @param _stride the step between data values (need *sizeof _type) e.g. 10*sizeof(GLfloat)
		/// @param _dataOffset Specifies a pointer to the first component of the first generic vertex attribute
		/// in the array. The initial value is 0. and is calculated as ((float *)NULL + (_dataOffset))
		/// @param _normalise specifies whether fixed-point data values should be normalized (GL_TRUE) or converted directly as fixed-point values
		/// (GL_FALSE) when they are accessed.
		//----------------------------------------------------------------------------------------------------------------------
		void setVertexAttributePointer(
																	GLuint _id,
																	GLint _size,
																	GLenum _type,
																	GLsizei _stride,
																	unsigned int _dataOffset,
																	bool _normalise=GL_FALSE
																);
	//----------------------------------------------------------------------------------------------------------------------
	/// @brief draw the VOA
	//----------------------------------------------------------------------------------------------------------------------
	void draw() const;
	//----------------------------------------------------------------------------------------------------------------------
	/// @brief set the number of faces to draw
	/// @param _n the number of indices to draw in glDrawArray (param 3 count)
	//----------------------------------------------------------------------------------------------------------------------
	inline void setNumIndices(GLuint _n){m_indicesCount=_n;}
protected :
	//----------------------------------------------------------------------------------------------------------------------
	/// @brief the draw mode of the VAO e.g. GL_TRIANGLES
	//----------------------------------------------------------------------------------------------------------------------
	GLenum m_drawMode;
	//----------------------------------------------------------------------------------------------------------------------
	/// @brief the number of faces to draw
	//----------------------------------------------------------------------------------------------------------------------
	GLenum m_indicesCount;
	//----------------------------------------------------------------------------------------------------------------------
	/// @brief the id of the VAO allocated by OpenGL
	//----------------------------------------------------------------------------------------------------------------------
	GLuint m_id;
	//----------------------------------------------------------------------------------------------------------------------
	/// @brief flag to indicate if we have allocated the data to the VAO
	//----------------------------------------------------------------------------------------------------------------------
	bool m_allocated;

	//----------------------------------------------------------------------------------------------------------------------
	/// @brief flag to indicate if we are bound
	//----------------------------------------------------------------------------------------------------------------------
	bool m_bound;
	//----------------------------------------------------------------------------------------------------------------------
	/// @brief hide the ctor as we want to create static factory only
	//----------------------------------------------------------------------------------------------------------------------
	VertexArrayObject(GLenum _mode);
	//----------------------------------------------------------------------------------------------------------------------
	/// @brief hide the copy ctor as we want to create static factory only
	//----------------------------------------------------------------------------------------------------------------------
	VertexArrayObject(const VertexArrayObject &_v);
	//----------------------------------------------------------------------------------------------------------------------
	/// @brief hide the dtor as we want to create static factory only
	//----------------------------------------------------------------------------------------------------------------------
	~VertexArrayObject(){;}

};


}
#endif

//----------------------------------------------------------------------------------------------------------------------
