/*
  Copyright (C) 2010 Jon Macey

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "NCCAPointBake.h"

//----------------------------------------------------------------------------------------------------------------------
/// @file NCCAPointBake.cpp
/// @brief implementation files for NCCAPointBake class
//----------------------------------------------------------------------------------------------------------------------

namespace ngl
{

//----------------------------------------------------------------------------------------------------------------------
NCCAPointBake::NCCAPointBake()
{
	m_numFrames=0;
	m_currFrame=0;
	m_nVerts=0;
	m_reader=0;
	m_startFrame=0;
	m_endFrame=0;
	m_mesh=0;
	m_reorderSize=0;
	m_binFile=false;



}



bool NCCAPointBake::loadPointBake(
                                  const std::string &_fileName
                                 )

{
  QFile toRead(_fileName.c_str());
  if (!toRead.open(QIODevice::ReadOnly | QIODevice::Text))
  {
    qDebug()<<"Could not open file";
    return false;
  }

  QXmlStreamReader xml(&toRead);
  unsigned int CurrentFrame=0;
  while(!xml.atEnd() && !xml.hasError())
  {
      /* Read next element.*/
      QXmlStreamReader::TokenType token = xml.readNext();
      /* If token is just StartDocument, we'll go to next.*/
      if(token == QXmlStreamReader::StartDocument)
      {
        continue;
      }

      /* If token is StartElement, we'll see if we can read it.*/
      else if(token == QXmlStreamReader::StartElement)
      {
        /* If it's named persons, we'll go to the next.*/
        if(xml.name() == "MeshName")
        {
          m_meshName=xml.readElementText().toStdString();
          qDebug()<<m_meshName.c_str();
        }

        else if(xml.name() == "NumVerts")
        {
          m_nVerts=xml.readElementText().toInt();
          qDebug()<<"Num verts "<<m_nVerts;
        }
        else if(xml.name() == "StartFrame")
        {
          m_startFrame=xml.readElementText().toInt();
          qDebug()<<"StartFrame "<<m_startFrame;
        }

        else if(xml.name() == "EndFrame")
        {
          m_endFrame=xml.readElementText().toInt();
          qDebug()<<"EndFrame "<<m_endFrame;

        }

        else if(xml.name() == "NumFrames")
        {
          m_numFrames=xml.readElementText().toInt();
          qDebug()<<"NumFrames "<<m_numFrames;
          //TOCHECK
          //first allocate base pointer [vertex]
          m_data.resize(m_numFrames);
          //cout <<"Size is now"<<m_data.size()<<endl;
          //now for each of these we need to allocate more space
          for(unsigned int i=0; i<m_numFrames; ++i)
          {
            m_data[i].resize(m_nVerts);
          }

        }
        else if(xml.name() == "TranslateMode")
        {
          qDebug()<<"found trans mode";

        }

        else if(xml.name() == "Vertex")
        {
          QXmlStreamAttributes attribs= xml.attributes();
          // vertex has an attribute called number which is used as the index
          // into the data array
          int index=attribs.value("number").toString().toInt();
          QString data=xml.readElementText();
          QStringList elements=data.split(" ");
          if(elements.size() < 3)
          {
            qDebug()<<"Vertex format error ";
          }
          else
          {
            m_data[CurrentFrame][index].set(elements.at(1).toFloat(),elements.at(2).toFloat(),elements.at(3).toFloat());

          }
        }
        else if(xml.name() == "Frame")
        {
          QXmlStreamAttributes attribs= xml.attributes();
          CurrentFrame=attribs.value("number").toString().toInt();
          CurrentFrame-=m_startFrame;
          std::cout<<"Found next Frame "<<CurrentFrame<<"\r";
					std::flush(std::cout);
        }
        else
        {
          //qDebug()<<"not sure what this is "<<xml.name();

        }
      }

   }
  return true;
}


//----------------------------------------------------------------------------------------------------------------------
NCCAPointBake::~NCCAPointBake()
{

}
//----------------------------------------------------------------------------------------------------------------------
NCCAPointBake::NCCAPointBake(
                             const std::string &_fileName
                            )
{

  loadPointBake(_fileName);
}
//----------------------------------------------------------------------------------------------------------------------
void NCCAPointBake::setFrame(
                              const unsigned int _frame
                            )
{
 m_currFrame=_frame;
}

bool NCCAPointBake::loadBinaryPointBake(
                                        const std::string &_fileName
                                       )
{
  // open a file stream for ip in binary mode
  std::fstream file;
  file.open(_fileName.c_str(),std::ios::in | std::ios::binary);
  // see if it worked
  if (!file.is_open())
  {
    std::cerr<<"problems Opening File "<<_fileName<<std::endl;
    return false;
  }
  // lets read in the header and see if the file is valid
  char header[11];
  file.read(header,10*sizeof(char));
  header[10]=0; // for strcmp we need \n
  // basically I used the magick string ngl::bin (I presume unique in files!) and
  // we test against it.
  if(strcmp(header,"ngl::binpb"))
  {
    // best close the file and exit
    file.close();
    std::cout<<"this is not an ngl::binpb file "<<std::endl;
    return false;
  }

  /// The number of vertices in the object
  //file.read(reinterpret_cast <char *> (&m_nVerts),sizeof(unsigned long int));

  file.read(reinterpret_cast <char *>(&m_numFrames),sizeof(unsigned int));
  file.read(reinterpret_cast <char *>(&m_currFrame),sizeof(unsigned int));
  file.read(reinterpret_cast <char *>(&m_nVerts),sizeof(unsigned int));
  file.read(reinterpret_cast <char *>(&m_startFrame),sizeof(unsigned int));
  file.read(reinterpret_cast <char *>(&m_reorderSize),sizeof(unsigned int));
  file.read(reinterpret_cast <char *>(&m_binFile),sizeof(bool));
  std::cout <<"Loaded header\n";
  std::cout<<m_numFrames<<"\n";
  std::cout<<m_currFrame<<"\n";
  std::cout<<m_nVerts<<"\n";
  std::cout<<m_startFrame<<"\n";
  std::cout<<m_reorderSize<<"\n";
  std::cout<<m_binFile<<"\n";

  m_dataReordered.resize(m_numFrames);
  m_data.resize(m_numFrames);
  std::cout<<"resided doing frames\n";
  for(unsigned int frame=0; frame<m_numFrames; ++frame)
  {
    m_dataReordered.resize(m_nVerts);
    m_data.resize(m_reorderSize);
  }
  std::cout<<"done re-size\n";
  for(unsigned int frame =0; frame<m_numFrames; ++frame)
  {
    m_data[frame].resize(m_nVerts);
    for(unsigned int v=0; v< m_nVerts; ++v)
    {
     file.read( reinterpret_cast <char *>(&m_data[frame][v].m_x),sizeof(float));
     file.read( reinterpret_cast <char *>(&m_data[frame][v].m_y),sizeof(float));
     file.read( reinterpret_cast <char *>(&m_data[frame][v].m_z),sizeof(float));
   }
  }
  std::cout<<"done m_data\n";

  for(unsigned int frame =0; frame<m_numFrames; ++frame)
  {
    m_dataReordered[frame].resize(m_reorderSize);

    for(unsigned int v=0; v< m_reorderSize; ++v)
    {
     file.read( reinterpret_cast <char *>(&m_dataReordered[frame][v].m_x),sizeof(float));
     file.read( reinterpret_cast <char *>(&m_dataReordered[frame][v].m_y),sizeof(float));
     file.read( reinterpret_cast <char *>(&m_dataReordered[frame][v].m_z),sizeof(float));
   }
  }

  std::cout<<"finished\n";
  return true;
}






bool NCCAPointBake::saveBinaryPointBake(
                                        const std::string &_fileName
                                       )
{
  // so basically we need to save all the state data from the abstract mesh
  // then map the vbo on the gpu and dump that in one go, this means we have to
  // call CreateVBO first the Save
    std::fstream file;
    file.open(_fileName.c_str(),std::ios::out | std::ios::binary);
    if (!file.is_open())
    {
      std::cerr<<"problems Opening File "<<_fileName<<std::endl;
      return false;
    }

    // lets write out our own Magic Number file ID
    const std::string header("ngl::binpb");
    file.write(header.c_str(),header.length());
    m_binFile=true;
    file.write(reinterpret_cast <char *>(&m_numFrames),sizeof(unsigned int));
    file.write(reinterpret_cast <char *>(&m_currFrame),sizeof(unsigned int));
    file.write(reinterpret_cast <char *>(&m_nVerts),sizeof(unsigned int));
    file.write(reinterpret_cast <char *>(&m_startFrame),sizeof(unsigned int));
    file.write(reinterpret_cast <char *>(&m_reorderSize),sizeof(unsigned int));
    file.write(reinterpret_cast <char *>(&m_binFile),sizeof(bool));
    std::cout<<m_numFrames<<"\n";
    std::cout<<m_currFrame<<"\n";
    std::cout<<m_nVerts<<"\n";
    std::cout<<m_startFrame<<"\n";
    std::cout<<m_reorderSize<<"\n";
    std::cout<<m_binFile<<"\n";

    // now write out data
    for(unsigned int frame =0; frame<m_numFrames; ++frame)
    {
      for(unsigned int v=0; v< m_nVerts; ++v)
      {
       file.write( reinterpret_cast <char *>(&m_data[frame][v].m_x),sizeof(float));
       file.write( reinterpret_cast <char *>(&m_data[frame][v].m_y),sizeof(float));
       file.write( reinterpret_cast <char *>(&m_data[frame][v].m_z),sizeof(float));
     }
    }

    for(unsigned int frame =0; frame<m_numFrames; ++frame)
    {
      for(unsigned int v=0; v< m_reorderSize; ++v)
      {
       file.write( reinterpret_cast <char *>(&m_dataReordered[frame][v].m_x),sizeof(float));
       file.write( reinterpret_cast <char *>(&m_dataReordered[frame][v].m_y),sizeof(float));
       file.write( reinterpret_cast <char *>(&m_dataReordered[frame][v].m_z),sizeof(float));
     }
    }

//    std::vector < std::vector<ngl::Vector> > m_data;
//    std::vector < std::vector<ngl::Vector> > m_dataReordered;



    file.close();
    return true;
}

void NCCAPointBake::setMeshToFrame(
                                   const unsigned int _frame
                                  )
  {
    // map the m_obj's vbo dat
    Real *ptr=m_mesh->mapVBOVerts();
    unsigned int step=0;
    // loop for all the verts and set the new vert value
    for(unsigned int i=0; i<m_reorderSize; ++i)
    {
      // grab the re-ordered index value
      // set it in the vbo pointer m_data
      ptr[step]=  m_dataReordered[_frame][i].m_x;
      ptr[step+1]=m_dataReordered[_frame][i].m_y;
      ptr[step+2]=m_dataReordered[_frame][i].m_z;
      // skip to the next set of 3 vert values

      step+=3;
    }
    // unmap the vbo as we have finished updating
    m_mesh->unMapVBO();
    m_currFrame=_frame;
 }


// as the m_obj class re-orders the vertices to pack them for the VBO
// we need to do the same for the clip
//----------------------------------------------------------------------------------------------------------------------
void NCCAPointBake::reorderVerts()
{
	// grab the re-ordered indices from the m_obj
	std::vector<IndexRef>outIndices=m_mesh->getIndices();
	// resize our m_data to the correct size
	unsigned int size=outIndices.size();
	m_dataReordered.resize(size);
  m_reorderSize=size;
	// now copy the indices from the vbo
	// m_data[CurrentFrame][index]
  // std::vector < std::vector<ngl::Vector> > m_dataReordered;

  // first we re-size the data to the VBO pack size
  std::cout<<"re-ordering verts first alloc space\n";
  m_dataReordered.resize(m_numFrames);
  std::cout <<"Reordered size = "<<size<<"\n";
  std::cout <<"Original Vets size = "<<m_mesh->m_nVerts<<"\n";

  for(unsigned int i=0; i<m_numFrames; ++i)
  {
    m_dataReordered[i].resize(size);
  }
  for( unsigned int frame=0; frame<m_numFrames; ++frame)
  {
    //std::cout<<"setting frame "<<frame<<"\n";

    for(unsigned int i=0; i<size; ++i)
    {
      m_dataReordered[frame][i].set(m_data[frame][outIndices[i].m_v][0],
                                    m_data[frame][outIndices[i].m_v][1],
                                    m_data[frame][outIndices[i].m_v][2],
                                    1.0
                                    );


    }
  }

}


//----------------------------------------------------------------------------------------------------------------------
bool NCCAPointBake::attachMesh(ngl::AbstractMesh *_mesh)
{

  if(_mesh->m_nVerts != m_nVerts)
  {
    std::cerr <<" Mesh can't be attached to this data as vert count does not match\n";
    std::cerr<<"mesh verts "<<_mesh->m_nVerts<<" file verts "<<m_nVerts<<"\n";
    return false;
  }

  else
  {
    m_mesh=_mesh;
    if(!m_binFile)
    {
      reorderVerts();
    }
    return true;
  }

}


//----------------------------------------------------------------------------------------------------------------------
std::vector<ngl::Vector> & NCCAPointBake::getRawDataPointerAtFrame(unsigned int _f)
{
	NGL_ASSERT( _f<=m_numFrames);
	return m_data[_f];

}

//----------------------------------------------------------------------------------------------------------------------
std::vector<ngl::Vector> & NCCAPointBake::getReorderedDataPointerAtFrame(unsigned int _f)
{
	NGL_ASSERT(_f<= m_numFrames);
	return m_dataReordered[_f];
}


} // end ngl namespace
//----------------------------------------------------------------------------------------------------------------------
